#ifndef _FILE_OFFSET_BITS
#define _FILE_OFFSET_BITS	64
#endif

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mount.h>

#include <uuid/uuid.h>

#include "endian.h"

#include "../struct.h"

static uuid_t uuid;

static struct timeval tv;

static unsigned long page_size;
static unsigned long inode_size;
static const char *device;

static uint64_t num_inodes;
static uint64_t bytes_per_inode;
static uint8_t log2_pages_per_group;

static uint64_t n_groups;
static uint64_t n_inodes;
static uint64_t inodes_in_group;
static uint64_t inodes_in_last_group;
static uint64_t inode_bitmap_inodes;

static int h;
static struct stat st;
static uint64_t size, n_pages;
static void *pg;
static struct nvfs_sb *sb;
static uint64_t *free_data;
static uint64_t dir_page;

static void set_bit(void *data, uint64_t bit)
{
	((uint8_t *)data)[bit >> 3] |= 1 << (bit & 7);
}

static bool test_bit(void *data, uint64_t bit)
{
	return (((uint8_t *)data)[bit >> 3] >> (bit & 7)) & 1;
}

static void wr_data(void *data, uint64_t offs, uint64_t len)
{
	ssize_t r;
	r = pwrite(h, data, len, offs);
	if (r == -1) {
		perror("mknvfs: pwrite");
		exit(1);
	}
}

static void wr_pg(void *data, uint64_t page, uint64_t n_pages)
{
	wr_data(data, page * page_size, n_pages * page_size);
}

static void calculate_inodes(void)
{
	for (sb->log2_inode_pages_per_group = 0; sb->log2_inode_pages_per_group < log2_pages_per_group; sb->log2_inode_pages_per_group++) {
		nvfs_get_sizes(sb,
				&n_groups,
				&n_inodes,
				&inodes_in_group,
				&inodes_in_last_group,
				&inode_bitmap_inodes);
		/*printf("n_groups: %lu\n", n_groups);
		printf("inodes %lu, wanted %lu\n", n_inodes, num_inodes);*/
		if (n_inodes >= num_inodes)
			goto ret;
	}
	sb->log2_inode_pages_per_group = log2_pages_per_group - 1;
ret:
	sb->root_inode = cpu_to_le64(inode_bitmap_inodes);
}

static void write_bitmaps(void)
{
	uint64_t i, j;
	for (i = 0; i < n_groups; i++) {
		uint64_t free_bits = page_size * 8;
		memset(pg, 0, page_size);
		if (!i)
			set_bit(pg, 0), free_bits--;
		set_bit(pg, 1), free_bits--;
		for (j = 0; j < 1U << sb->log2_inode_pages_per_group; j++) {
			set_bit(pg, 2 + j), free_bits--;
		}
		if (!i) {
			set_bit(pg, 3 + j), free_bits--;
			dir_page = 3 + j;
		}
		if (i == n_groups - 1) {
			j = (1U << log2_pages_per_group) - 1;
			while ((i << log2_pages_per_group) + j >= n_pages) {
				if (test_bit(pg, j)) {
					fprintf(stderr, "mknvfs: the device is too small\n");
					exit(1);
				}
				set_bit(pg, j), free_bits--;
				j--;
			}
		}
		wr_pg(pg, 1 + (i << log2_pages_per_group), 1);
		free_data[i] = free_bits;
	}
}

static void write_inodes(void)
{
	uint64_t i, j;
	struct nvfs_group *nvg;
	uint64_t sz = (inode_bitmap_inodes + inodes_in_group) << sb->log2_inode_size;
	sz = (sz + page_size - 1) & -(uint64_t)page_size;
	nvg = valloc(sz);
	if (!nvg) {
		perror("valloc");
		exit(1);
	}
	for (i = 0; i < n_groups; i++) {
		uint64_t inodes;
		memset(nvg, 0, sz);
		for (j = 0; j < inode_bitmap_inodes; j++)
			set_bit(nvg->bitmap, j);
		if (!i) {
			struct nvfs_inode *nvi = (struct nvfs_inode *)((uint8_t *)nvg + (inode_bitmap_inodes << sb->log2_inode_size));
			set_bit(nvg->bitmap, j);
			nvi->i_size = cpu_to_le64(page_size);
			nvi->i_blocks = cpu_to_le64(page_size / 512);
			for (j = 0; j < NVFS_RADIX_BLOCKS; j++)
				nvi->i_radix[0][j] = cpu_to_le64(dir_page * page_size);
			nvi->i_ctime = nvi->i_mtime = nvi->i_atime = cpu_to_le64(tv.tv_sec);
			nvi->i_ctime_nsec = nvi->i_mtime_nsec = nvi->i_atime_nsec = cpu_to_le32(tv.tv_usec * 1000);
			nvi->i_links = cpu_to_le32(2);
			nvi->i_mode = cpu_to_le32(0755 | S_IFDIR);
			nvi->i_xattr[0].x = nvfs_xattr_entry_make_size(inode_size - offsetof(struct nvfs_inode, i_xattr));
		}
		inodes = i == n_groups - 1 ? inodes_in_last_group : inodes_in_group;
#if 0
		for (j = inode_bitmap_inodes + inodes; j / 8 + sizeof(struct nvfs_group) < inode_bitmap_inodes << sb->log2_inode_size; j++) {
			set_bit(nvg->bitmap, j);
		}
#endif
		nvg->n_free_blocks = cpu_to_le32(free_data[i]);
		nvg->n_free_inodes = cpu_to_le32(inodes - !i);
		wr_pg(nvg, 2 + (i << log2_pages_per_group), sz / page_size);
	}
	free(nvg);
}

static void write_root_dir(void)
{
	memset(pg, 0, page_size);
	((struct nvfs_dir_head *)pg)->magic = NVFS_DIR_HEAD_MAGIC;
	((struct nvfs_dir_head *)pg)->self = cpu_to_le64(dir_page * page_size);
	((struct nvfs_dir_head *)pg)->next = cpu_to_le64(0);
	((struct nvfs_dir_head *)pg)->inode = sb->root_inode;
	((struct nvfs_dir_head *)pg)->nde->inode = cpu_to_le64(0);
	((struct nvfs_dir_head *)pg)->nde->name_len_and_type = cpu_to_le32(nvfs_dir_entry_make_size(page_size - NVFS_DIR_OFFSET) << NAME_LEN_OFFSET);
	wr_pg(pg, dir_page, 1);
}

int main(int argc, char *argv[])
{
	int a;
	bool no_args;
	char *end;
	unsigned i;
	uint8_t checksum;

	uuid_generate(uuid);

	if (gettimeofday(&tv, NULL)) {
		perror("mknvfs: gettimeofday");
		exit(1);
	}

	page_size = getpagesize();
	if (page_size > 1U << MAX_LOG_PAGE_SIZE) {
		fprintf(stderr, "mknvfs: system page size is greater than 64k\n");
		exit(1);
	}
	inode_size = 1U << MIN_LOG_INODE_SIZE;
	device = NULL;
	num_inodes = 0;
	bytes_per_inode = 0;

	no_args = false;
	for (a = 1; a < argc; a++) {
		const char *str = argv[a];
		if (!strcmp(str, "--")) {
			no_args = true;
			continue;
		}
		if (!no_args && str[0] == '-') {
			if (a == argc - 1) {
				fprintf(stderr, "mknvfs: invalid arguments\n");
				exit(1);
			}
			if (!strcmp(str, "-b")) {
				page_size = strtoul(argv[++a], &end, 10);
				if (*end ||
				    page_size & (page_size - 1) ||
				    page_size < 1U << MIN_LOG_PAGE_SIZE ||
				    page_size > 1U << MAX_LOG_PAGE_SIZE) {
					fprintf(stderr, "mknvfs: invalid page size\n");
					exit(1);
				}
				continue;
			}
			if (!strcmp(str, "-I")) {
				inode_size = strtoul(argv[++a], &end, 10);
				if (*end ||
				    inode_size & (inode_size - 1) ||
				    inode_size < 1U << MIN_LOG_INODE_SIZE) {
invalid_inode_size:
					fprintf(stderr, "mknvfs: invalid inode size\n");
					exit(1);
				}
				continue;
			}
			if (!strcmp(str, "-N")) {
				num_inodes = strtoull(argv[++a], &end, 10);
				if (*end || !num_inodes) {
					fprintf(stderr, "mknvfs: invalid number of inodes\n");
					exit(1);
				}
				continue;
			}
			if (!strcmp(str, "-i")) {
				bytes_per_inode = strtoull(argv[++a], &end, 10);
				if (*end || !bytes_per_inode) {
					fprintf(stderr, "mknvfs: invalid number of bytes per inodes\n");
					exit(1);
				}
				continue;
			}
			goto invalid_arguments;
		}
		if (!device) {
			device = str;
		} else {
invalid_arguments:
			fprintf(stderr, "mknvfs: invalid arguments\n");
			exit(1);
		}
	}
	if (inode_size > page_size)
		goto invalid_inode_size;
	if (!device)
		goto invalid_arguments;

	h = open(device, O_RDWR);
	if (h == -1) {
		perror("mknvfs: open");
		exit(1);
	}

	if (fstat(h, &st)) {
		perror("mknvfs: fstat");
		exit(1);
	}
	if (S_ISBLK(st.st_mode)) {
		if (ioctl(h, BLKGETSIZE64, &size)) {
			perror("mknvfs: ioctl BLKGETSIZE64");
			exit(1);
		}
	} else if (S_ISREG(st.st_mode)) {
		size = st.st_size;
	} else {
		fprintf(stderr, "mknvfs: invalid file type\n");
		exit(1);
	}
	n_pages = size / page_size;
	if (!n_pages) {
too_small:
		fprintf(stderr, "mknvfs: the device is too small\n");
		exit(1);
	}
	n_pages--;

	if (!num_inodes) {
		if (!bytes_per_inode) {
			bytes_per_inode = DEFAULT_BYTES_PER_INODE;
			if (bytes_per_inode < page_size)
				bytes_per_inode = page_size;
		}
		num_inodes = size / bytes_per_inode;
	} else {
		if (bytes_per_inode) {
			fprintf(stderr, "mknvfs: -N and -i can't be used simultaneously\n");
			exit(1);
		}
	}
	if (num_inodes > n_pages)
		num_inodes = n_pages;
	if (!num_inodes)
		goto too_small;

	sb = valloc(page_size);
	if (!sb) {
		perror("mknvfs: valloc");
		exit(1);
	}
	memset(sb, 0, page_size);

	pg = valloc(page_size);
	if (!pg) {
		perror("mknvfs: valloc");
		exit(1);
	}

	memcpy(sb->magic, NVFS_MAGIC, NVFS_MAGIC_LEN);
	memcpy(sb->uuid, uuid, NVFS_UUID_LEN);
	sb->n_pages = cpu_to_le64(n_pages);
	sb->log2_page_size = ffs(page_size) - 1;
	sb->log2_inode_size = ffs(inode_size) - 1;

	log2_pages_per_group = sb->log2_page_size + 3;

	calculate_inodes();

	free_data = malloc(n_groups * sizeof(uint64_t));
	if (!free_data) {
		perror("mknvfs: malloc");
		exit(1);
	}

	write_bitmaps();
	write_inodes();
	write_root_dir();

	checksum = 0;
	for (i = 0; i < NVFS_SB_STATIC_PART; i++)
		checksum += ((uint8_t *)sb)[i];
	sb->checksum = -checksum;

	wr_pg(sb, 0, 1);
	wr_pg(sb, n_pages, 1);

	if (fsync(h)) {
		perror("mknvfs: fsync");
		exit(1);
	}
	if (close(h)) {
		perror("mknvfs: close");
		exit(1);
	}
	free(sb);
	free(pg);
	free(free_data);

	return 0;
}
